/* $Header: /src/carson/nistnet/RCS/knistnet.c,v 1.5 1998/01/30 17:25:32 carson Exp $ */

/* knistnet.c - Linux implementation of "hitbox"-like functionality.
 * This code exists as a loadable kernel module.  It gains access to the
 * entry points it needs through some patches to the existing Linux kernel.
 * (Unfortunately, there seemed to be no practical alternative to patching
 * to gain access to the basic packet handling routine entry points.)
 * The user-level interface provided is a device driver one, modeled on the
 * original SunOS-based hitbox.
 */

#include "kincludes.h"
#include <linux/version.h>
#include <linux/module.h>

#include "tabledist.h"

int domunge;
int hit_debug;
#ifndef MODULE
#define BREAKPOINT() asm("   int $3")
#else
#define BREAKPOINT()
#endif

int LinLockVar;

#define LinLock() \
		{\
		save_flags(pre_flags);\
		cli();\
		if (LinLockVar) {\
			BREAKPOINT();\
		}\
		++LinLockVar;\
		}
#define LinUnlock() \
		{\
		if (LinLockVar != 1) {\
			BREAKPOINT();\
		}\
		--LinLockVar;\
		restore_flags(pre_flags);\
		}


#define HASHSIZE 256
static struct lin_hitbox *hittable[HASHSIZE];

#ifdef DEBUG
static int hittable_count;
#endif

struct lin_globalstats ourstats;
#define	STATS_START	0
#define	STATS_PROCESS	1
#define STATS_UNPROCESS	2

void lin_hash_stats(int number);

void re_fast_fill(void);
void fast_fill(int how);
void fast_empty(void);
struct fast_timer_list * fast_alloc(int how);
void fast_free(struct fast_timer_list *done);

void
freehittable(void)
{
	struct lin_hitbox *freeme, *frome;
	int i;
#ifdef DEBUG
	int count=0;
#endif

	for (i=0; i < HASHSIZE; ++i) {
		freeme = hittable[i];
		while (freeme) {
			frome = freeme->next;
			kfree_s(freeme, sizeof(struct lin_hitbox));
			freeme = frome;
#ifdef DEBUG
			++count;
#endif
		}
	}
#ifdef DEBUG
	if (count != hittable_count) {
		printk("Linhit: expected %d, found %d\n",
			hittable_count, count);
	}
#endif
}


/* Hash an entry by src/dest address */
int
hithash(u_int32_t src, u_int32_t dest)
{
	unsigned int value, answer=0;

	/* this is not intended to be an extraordinarily good hash! */
	value = (unsigned int)src*69069 + (unsigned int)dest;
	while (value) {
		answer |= value;
		value >>= 8;
	}
	return (int) (answer&(HASHSIZE-1));
}

struct lin_hitbox *
findhit(u_int32_t src, u_int32_t dest)
{
	register struct lin_hitbox *answer;
	int hash_tries = 0;
	
	answer = hittable[hithash(src, dest)];
	while (answer) {
		++hash_tries;
		if (answer->hit_entry.src == src
			&& answer->hit_entry.dest == dest) {
			lin_hash_stats(hash_tries);
			return answer;
		}
		answer = answer->next;
	}

	++hash_tries;
	answer = hittable[hithash(src, INADDR_ANY)];
	while (answer) {
		++hash_tries;
		if (answer->hit_entry.src == src
			&& answer->hit_entry.dest == INADDR_ANY) {
			lin_hash_stats(hash_tries);
			return answer;
		}
		answer = answer->next;
	}

	++hash_tries;
	/* Try again with generic addresses */
	answer = hittable[hithash(INADDR_ANY, dest)];
	while (answer) {
		++hash_tries;
		if (answer->hit_entry.src == INADDR_ANY
			&& answer->hit_entry.dest == dest) {
			lin_hash_stats(hash_tries);
			return answer;
		}
		answer = answer->next;
	}
if (hit_debug) {
	unsigned char *srcp, *destp;

	srcp = (unsigned char *)&src;
	destp = (unsigned char *)&dest;
	printk("hitbox: missed on %d.%d.%d.%d->%d.%d.%d.%d\n",
			srcp[0], srcp[1], srcp[2], srcp[3],
			destp[0], destp[1], destp[2], destp[3]);
}
	++hash_tries;
	answer = hittable[hithash(INADDR_ANY, INADDR_ANY)];
	while (answer) {
		++hash_tries;
		if (answer->hit_entry.src == INADDR_ANY
			&& answer->hit_entry.dest == INADDR_ANY) {
			lin_hash_stats(hash_tries);
			return answer;
		}
		answer = answer->next;
	}
	++hash_tries;
	lin_hash_stats(hash_tries);
	return NULL;
}

int
rmhitreq(struct lin_hitreq *hitreq)
{
	struct lin_hitbox *victim, *parent;
	
	victim = hittable[hithash(hitreq->src, hitreq->dest)];
	if (!victim) return -ESRCH;
	if (victim->hit_entry.src == hitreq->src &&
		victim->hit_entry.dest == hitreq->dest) {
			hittable[hithash(hitreq->src, hitreq->dest)]
						= victim->next;
			kfree_s(victim, sizeof(struct lin_hitbox));
#ifdef DEBUG
			--hittable_count;
#endif
	} else for (parent=victim, victim=victim->next; victim;
			parent=victim, victim=victim->next) {
		if (victim->hit_entry.src == hitreq->src &&
			victim->hit_entry.dest == hitreq->dest) {
			parent->next = victim->next;
			kfree_s(victim, sizeof(struct lin_hitbox));
#ifdef DEBUG
			--hittable_count;
#endif
			break;
		}
	}
	return 0;
}


int
addhitreq(struct lin_hitreq *hitreq)
{
	struct lin_hitbox *boxer;

	/* Change delays from usec to minijiffies */
	hitreq->delay = usec_to_minijiffy(hitreq->delay);
	hitreq->delsigma = usec_to_minijiffy(hitreq->delsigma);


	/* First see if we're already hitting on this one */
	boxer = findhit(hitreq->src, hitreq->dest);
	if (boxer && boxer->hit_entry.src == hitreq->src &&
		boxer->hit_entry.dest == hitreq->dest) {
		/* replace entry */
		boxer->hit_entry = *hitreq;
	} else {
		/* need a new one */
		boxer = kmalloc(sizeof(struct lin_hitbox), GFP_KERNEL);
		if (!boxer)
			return -ENOMEM;
		memset(boxer, 0, sizeof(struct lin_hitbox));
		boxer->hit_entry = *hitreq;
		/* Initialize packet timer */
		do_gettimeofday(&boxer->stats.last_packet);
		boxer->stats.next_packet = boxer->stats.last_packet;
		/* Insert in table */
		boxer->next = hittable[hithash(hitreq->src, hitreq->dest)];
		hittable[hithash(hitreq->src, hitreq->dest)] = boxer;
#ifdef DEBUG
			++hittable_count;
#endif
	}
	return 0;
}


/* nice and high, but it is tunable: insmod hitmod.o major=your_selection */
static int major = HITMAJOR;

/*
 * The driver.
 */

/* read -- get what's in the table */
static int hw_read(struct inode * node,struct file * file,char * buf,int count)
{
	int i, origcount = count;
	struct lin_hitbox *readme;
	struct addrpair out;

	for (i=0; i < HASHSIZE && count >= sizeof(struct addrpair); ++i) {
		readme = hittable[i];
		while (readme && count >= sizeof(struct addrpair)) {

			out.src = readme->hit_entry.src;
			out.dest = readme->hit_entry.dest;
			memcpy_tofs(buf, (char *)&out,
				(unsigned long) sizeof(struct addrpair));
			readme = readme->next;
			buf += sizeof(struct addrpair);
			count -= sizeof(struct addrpair);
		}
	}
#ifdef DEBUG
	if (origcount-count > hittable_count) {
		printk("Linhit: %d new guys showed up somehow\n",
			(origcount-count)-hittable_count);
	}
#endif
	return origcount-count;
}

static int hw_write(struct inode * node,struct file * file,const char * buf,int count)
{
	if (count >= tabledistsize()) {
		tabledistfill(buf);
		return tabledistsize();
	} else
		return -E2BIG;	/* ha, ha */
}

/*
 * mucky muck ioctl interface
 */
static int hw_ioctl(struct inode * inode, struct file * file, unsigned int type, unsigned long arg)
{
	struct lin_hitreq hitreq;
	int rc = 0;

	switch (type) {
	case HITIOCTL_OFF:
		/* Shut down */
		domunge=0;
		break;
	case HITIOCTL_ON:
		/* Turn on if not already on */
		domunge=1;
		break;
	case HITIOCTL_ADD:
		/* Get what they want */
		memcpy_fromfs(&hitreq, (struct lin_hitreq *)arg,
			sizeof(struct lin_hitreq));
		/* Add it to the table */
		rc = addhitreq(&hitreq);
		break;
	case HITIOCTL_REMOVE:
		/* Get what they want */
		memcpy_fromfs(&hitreq, (struct lin_hitreq *)arg,
			sizeof(struct lin_hitreq));
		/* Remove it from the table */
		rc = rmhitreq(&hitreq);
		break;
	case HITIOCTL_STATS:
		{
		struct lin_hitstats hitstats;
		struct lin_hitbox *where;
		int i;

		/* Get what they want */
		memcpy_fromfs(&hitstats, (struct lin_hitstats *)arg,
			sizeof(struct lin_hitstats));
		where = findhit(hitstats.hitreq.src, hitstats.hitreq.dest);
		if (!where)
			return -ESRCH;
		hitstats = where->stats;
		/* Change delays from minijiffies to usec */
		hitstats.hitreq.delay = minijiffy_to_usec(hitstats.hitreq.delay);
		hitstats.hitreq.delsigma = minijiffy_to_usec(hitstats.hitreq.delsigma);
		/* Compute current_bandwidth */
		for (i=0; i < BAND_ARRAY; ++i)
			hitstats.current_bandwidth +=
				hitstats.bandwidth_array[i];
		if (hitstats.seats_used)
			hitstats.current_bandwidth /= hitstats.seats_used;
		/* Copy out individual stats */
		memcpy_tofs((struct lin_hitstats *)arg,
			&hitstats,
			sizeof(struct lin_hitstats));
		break;
		}
	case HITIOCTL_MODE:
		/* ?? */
		break;
	case HITIOCTL_TIMER:
		/* ?? */
		break;
	case HITIOCTL_MTU:
		/* ?? */
		break;
	case HITIOCTL_DEBUG:
		hit_debug = get_user((long *)arg);
		break;
	case HITIOCTL_GLOBALSTATS:
		memcpy_tofs((struct lin_globalstats *)arg,
			&ourstats,
			sizeof(struct lin_globalstats));
		break;
	}
	return rc;
}


/*
 * Our special open code.
 * MOD_INC_USE_COUNT make sure that the driver memory is not freed
 * while the device is in use.
 */
static int
hw_open( struct inode* ino, struct file* filep)
{
	MOD_INC_USE_COUNT;
	return 0;   
}

/*
 * Now decrement the use count.
 */
static void
hw_close( struct inode* ino, struct file* filep)
{
	MOD_DEC_USE_COUNT;
}

static struct file_operations hw_fops = {
	NULL,
	hw_read,
	hw_write,
	NULL,		/* hw_readdir */
	NULL,		/* hw_select */
	hw_ioctl,		/* hw_ioctl */
	NULL,
	hw_open,
	hw_close,
	NULL		/* fsync */
};


/*
 * And now the modules code and kernel interface.
 */


#ifndef abs
#define abs(a) ((a)>=0?(a):-(a))
#endif

/* return a-b difference in microseconds */
static int
timeval_diff(struct timeval *a, struct timeval *b)
{
	int answer;

	if (a->tv_sec > b->tv_sec) {
		answer = (a->tv_sec - b->tv_sec)-1;
		answer *= MILLION;
		answer += a->tv_usec+MILLION - b->tv_usec;
		return answer;
	} else if (a->tv_sec == b->tv_sec) {
		return a->tv_usec - b->tv_usec;
	} else {
		answer = (b->tv_sec - a->tv_sec)-1;
		answer *= MILLION;
		answer += b->tv_usec+MILLION - a->tv_usec;
		return -answer;
	}
}

/* Return time difference a-b in minijiffies */
static int
timeval_minidiff(struct timeval *a, struct timeval *b)
{
	return usec_to_minijiffy(timeval_diff(a, b));
}

/* Add on some usec to time */
static void
timeval_add(struct timeval *a, unsigned long more_usec)
{
	a->tv_usec += more_usec;
	if (a->tv_usec > MILLION) {
		a->tv_usec -= MILLION;
		++a->tv_sec;
	}
}

#define band_seat(time) ((time.tv_sec)%BAND_ARRAY)


void
lin_hash_stats(int number)
{
	static int hash_slot;

	ourstats.hash_tries[hash_slot] += number;
	ourstats.hash_tries[hash_slot] >>= 1;
	hash_slot = (hash_slot+1)%BAND_ARRAY;
}

void
global_stats(int process)
{
	static struct timeval start;
	struct timeval end;
	static int process_slot, unprocess_slot;

	switch(process) {
	case STATS_START:
		do_gettimeofday(&start);
		return;
	case STATS_PROCESS:
		do_gettimeofday(&end);
		ourstats.process_overhead[process_slot] +=
			(unsigned long)abs(timeval_diff(&end, &start));
		ourstats.process_overhead[process_slot] >>= 1;
		process_slot = (process_slot+1)%BAND_ARRAY;
		break;
	case STATS_UNPROCESS:
		do_gettimeofday(&end);
		ourstats.unprocess_overhead[unprocess_slot] +=
			(unsigned long)abs(timeval_diff(&end, &start));
		ourstats.unprocess_overhead[unprocess_slot] >>= 1;
		unprocess_slot = (unprocess_slot+1)%BAND_ARRAY;
		break;
	}
}

void
packet_stats(struct sk_buff *skb, struct lin_hitbox *hitme)
{
	struct timeval our_time;
	unsigned long packettime;
	int last_seat, our_seat, seat;

	if (!skb->len)
		return;
	do_gettimeofday(&our_time);
	last_seat = band_seat(hitme->stats.last_packet);
	our_seat = band_seat(our_time);
	packettime = (unsigned long)abs(timeval_diff(&our_time, &hitme->stats.last_packet));
	hitme->stats.last_packet = our_time;
	/* compute bandwidth */
	if (packettime > BAND_ARRAY*MILLION) {
		/* too long since last packet; zero out everything */
		memset(hitme->stats.bandwidth_array, 0,
			BAND_ARRAY*sizeof(unsigned long));
		hitme->stats.seats_used = 1;
	} else if (last_seat != our_seat) {	/* zero out skipped intervals */
		for (seat = (last_seat+1)%BAND_ARRAY; seat != our_seat;
			seat = (seat+1)%BAND_ARRAY) {
				hitme->stats.bandwidth_array[seat] = 0;
		}
		/* plus get this one! */
		hitme->stats.bandwidth_array[seat] = 0;
		if (hitme->stats.seats_used < BAND_ARRAY)
			++hitme->stats.seats_used;
	}
	hitme->stats.bandwidth_array[our_seat] += skb->len;

	hitme->stats.current_size = skb->len;
	hitme->stats.bytes_sent += skb->len;
	return;
}

/* Receive packet interception */
static struct packet_type *ippt;

static struct packet_type ourpt;

void
runpacket(struct fast_timer_list *info)
{
	struct hit_packetinfo *hpi = (struct hit_packetinfo *)info->data;
unsigned long pre_flags;

LinLock();
	packet_stats(hpi->skb, hpi->hitme);
LinUnlock();
	restore_flags(hpi->flags);
	(void) ippt->func(hpi->skb, hpi->dev, ippt);
LinLock();
if (!hpi->hitme->stats.qlen) {
	BREAKPOINT();
}
	--hpi->hitme->stats.qlen;
	fast_free(info);
LinUnlock();
	MOD_DEC_USE_COUNT;
}

/* This is the (slightly modified) DRD algorithm for dropping */

/* probability factors * PROBFACTOR */
static unsigned int drdtable[] = {	/* constant "ramp up" */
	6554,
	9830,
	13107,
	16384,
	19661,
	22938,
	26214,
	29491,
	32768,
	36045,
	39322,
	42598,
	45875,
	49152,
	52429,
	55706,
	58982,
	62259,
	65535
};
#define DRDLIMIT (sizeof(drdtable)/sizeof(int))

/* dropprob - compute probability of dropping packet.  If we are using
 * DRD, use its table (adjusted for the queue length parameters we're
 * using), otherwise use a constant drop probability (which may be 0).
 */
int
dropprob(struct lin_hitbox *hitme, int *use_drd)
{
	if (hitme->hit_entry.drd_max) {	/* using DRD */
	    int stretch = DRDLIMIT/(hitme->hit_entry.drd_max-hitme->hit_entry.drd_min);

	    if (hitme->stats.qlen < hitme->hit_entry.drd_min) {
	    	*use_drd=0;
		return hitme->hit_entry.p_drop;
	    }
	    *use_drd=1;
	    if (hitme->stats.qlen >= hitme->hit_entry.drd_max)
		return drdtable[DRDLIMIT-1];
	    else
		return drdtable[stretch *
			(hitme->stats.qlen-hitme->hit_entry.drd_min)];
	} else {
		*use_drd=0;
		return hitme->hit_entry.p_drop;
	}
}


/* Determine the amount of time to delay a packet.  This is the maximum
 * of two quantities:
 *	1. Probabilistic packet delay time
 *	2. Bandwidth-limitation delay time
 *
 * Question: Should we take probabilistic delay into account in determining
 * bandwidth consumption?  Answer: This complicates things a little too much.
 * Our model is that bandwidth throttling happens first at one virtual
 * choke point, then packets may get independently delayed at some later
 * point.  This can result in packets getting bunched up, so the stated
 * bandwidth limitation is actually exceeded at some point.
 *
 * The problem with taking the delay into account is that the simplest
 * way of doing so would remove any possibility of reordering packets --
 * each packet could not be sent any sooner than its predecessor.  It
 * thus seems more useful in terms of network effects to do things the
 * way they are here.
 */
int
packet_delay(struct sk_buff *skb, struct lin_hitbox *hitme)
{
	int probdelay=0, bandwidthdelay=0;

	/* Figure probabilistic delay */
	if (hitme->hit_entry.delsigma)
		probdelay = tabledist(hitme->hit_entry.delay,
			hitme->hit_entry.delsigma);
	else
		probdelay = hitme->hit_entry.delay;

	/* Figure bandwidth-limitation delay */
	if (hitme->hit_entry.bandwidth) {
		struct timeval our_time;
		unsigned long packettime;

		do_gettimeofday(&our_time);
		/* We can't send until queued packets have been taken care of */
		bandwidthdelay = timeval_minidiff(&hitme->stats.next_packet,
					&our_time);
		if (bandwidthdelay < 0) {
			bandwidthdelay = 0;
			hitme->stats.next_packet = our_time;
		}
		/* Now determine how much time this packet will take in
		 * usec, in order to schedule the following one.
		 */
		/* skb->len is probably too big (with other junk) */
		packettime = (unsigned long)
			( ((double)skb->len*MILLION) / (double)hitme->hit_entry.bandwidth);
		timeval_add(&hitme->stats.next_packet, packettime);
#ifdef DEBUG
{
static int ticker;

++ticker;
if (!(ticker&0x3f)) {
	printk("Packet size %ld delay %ld usec\n",
		skb->len, packettime);
printk("Current time is %d.%06d, will send at %d.%06d\n",
		our_time.tv_sec, our_time.tv_usec,
		hitme->stats.next_packet.tv_sec, hitme->stats.next_packet.tv_usec);
}
}
#endif
	}


	return probdelay > bandwidthdelay ? probdelay : bandwidthdelay;
}

int
default_munger(struct sk_buff *skb, struct device *dev, struct packet_type *ptype, struct lin_hitbox *hitme)
{
	return 1;
}

packet_munger other_munge = default_munger;

void addmunge(packet_munger munger)
{
	other_munge = munger;
}

void rmmunge(packet_munger munger)
{
	if (munger == other_munge)
		other_munge = default_munger;
}

#define munge_finish()	{LinUnlock(); if (skb2) (void) rcv_packet_munge(skb2, dev, ptype);/* recursively process dup */ }

int
rcv_packet_munge(struct sk_buff *skb, struct device *dev, struct packet_type *ptype)
{
unsigned long pre_flags;

LinLock();
	if (domunge) {
		int maybedrop, use_drd, delaytime;
		struct lin_hitbox *hitme;
		struct fast_timer_list *screamer;
		struct hit_packetinfo *hpi;
		struct sk_buff *skb2=NULL;
		int ret;

		global_stats(STATS_START);
		hitme = findhit(skb->h.iph->saddr, skb->h.iph->daddr);
		ret = (*other_munge)(skb, dev, ippt, hitme);
		if (ret <= 0) {
			global_stats(STATS_PROCESS); /* well, sort of */
LinUnlock();
			return ret;
		}
		if (!hitme) {	/* not intercepting */
			global_stats(STATS_UNPROCESS);
LinUnlock();
			return ippt->func(skb, dev, ippt);
		}

		/* Assume we will queue until we find otherwise */
		++hitme->stats.qlen;

		/* See if we're going to drop the packet */
		if ((maybedrop = dropprob(hitme, &use_drd))) {
			int value = (random()&0xffff);

			if (value < maybedrop) {
				kfree_skb(skb, FREE_WRITE);
				--hitme->stats.qlen;
				++hitme->stats.n_drops;
				if (use_drd)
					++hitme->stats.drd_drops;
				else
					++hitme->stats.rand_drops;
				global_stats(STATS_PROCESS);
LinUnlock();
				return 0;
			}
		}

		/* See if we're going to duplicate the packet.  Here,
		 * we just do fixed probability.
		 */
		if (hitme->hit_entry.p_dup) {
			int value = (random()&0xffff);
			
			if (value < hitme->hit_entry.p_dup) {	/* you get a new sister! */

				++hitme->stats.dups;
				skb2 = skb_copy(skb, GFP_ATOMIC);
			}
		}

		/* Now see if we're going to delay the packet */
		if (!(delaytime = packet_delay(skb, hitme))) { /* no delay */
			--hitme->stats.qlen;
			packet_stats(skb, hitme);

			global_stats(STATS_PROCESS);
munge_finish();
			return ippt->func(skb, dev, ippt);
		}

		screamer = fast_alloc(GFP_ATOMIC);
		/* If we can't allocate, punt! */
		if (!screamer) {
			--hitme->stats.qlen;
			++hitme->stats.mem_drops;
			packet_stats(skb, hitme);
			global_stats(STATS_PROCESS);
munge_finish();
			return ippt->func(skb, dev, ippt);
		}
		hpi = (struct hit_packetinfo *)screamer->data;

		init_fast_timer(screamer);
		hpi->skb = skb;
		hpi->dev = dev;
		hpi->hitme = hitme;
		hpi->flags = pre_flags;

		/* Schedule something to happen in a little while */
		screamer->expires = delaytime;
		MOD_INC_USE_COUNT;
		add_fast_timer(screamer);
		global_stats(STATS_PROCESS);
munge_finish();
		return 0;
	} else {
LinUnlock();
		return ippt->func(skb, dev, ippt);
	}
}

void
grab_ip_rcv(void)
{
	ippt = dev_find_pack(htons(ETH_P_IP));
	if (ippt) {
		ourpt = *ippt;
		ourpt.func = rcv_packet_munge;
		dev_remove_pack(ippt);
		dev_add_pack(&ourpt);
	}
}

void
release_ip_rcv(void)
{
	if (ippt) {
		dev_remove_pack(&ourpt);
		dev_add_pack(ippt);
		ippt = NULL;
	}
}

#ifdef MODULE
int
init_module( void)
#else
int
nistnet_init(void)
#endif
{
	void export_nistnet_symbols(void);

	printk( "nistnet: init_module called\n");
	if (register_chrdev(major, "hw", &hw_fops)) {
		printk("nistnet: register_chrdev failed: goodbye world :-(\n");
		return -EIO;
	}
	else
		printk("nistnet: Hello, world\n");

	re_fast_fill();
	memset(&ourstats, 0, sizeof(ourstats));
	grab_ip_rcv();
	export_nistnet_symbols();
	return 0;
}

#ifdef MODULE
void
cleanup_module(void)
{
	printk("nistnet:  cleanup_module called\n");

	if (unregister_chrdev(major, "hw") != 0)
		printk("nistnet: cleanup_module failed\n");
	else
		printk("nistnet: cleanup_module succeeded\n");
	domunge = 0;
	release_ip_rcv();
	freehittable();
	fast_empty();
}
#endif

/* This isn't quite right for allowing extra refills, since I still
 * have a possible race condition... */

struct fast_timer_list *fast_stack;

#define FAST_RESERVE 1000

struct fast_timer_list *bigfastspace;
struct hit_packetinfo *bighpispace;

void
fast_fill(int how)
{
	struct fast_timer_list *newfast, *fastspace;
	struct hit_packetinfo *newhpi, *hpispace;
	int i, limit;

	if (how == GFP_ATOMIC) {	/* @@ need to track for free...*/
		limit = 1;
		fastspace = (struct fast_timer_list *)
			kmalloc(sizeof(struct fast_timer_list), how);
		hpispace = (struct hit_packetinfo *)
			kmalloc(sizeof(struct hit_packetinfo), how);
	} else {
		limit = FAST_RESERVE;
		bigfastspace = fastspace = (struct fast_timer_list *)
			vmalloc(sizeof(struct fast_timer_list)*limit);
		bighpispace = hpispace = (struct hit_packetinfo *)
			vmalloc(sizeof(struct hit_packetinfo)*limit);
	}
		
	for (i=0; i < limit; ++i) {
		newfast = fastspace+i;
		newhpi = hpispace+i;
		newfast->data = (unsigned long) newhpi;
		newfast->function = runpacket;
		newfast->next = fast_stack;
		fast_stack = newfast;
	}
}

void
fast_empty(void)
{
	/*struct fast_timer_list *onward;*/

	vfree((void *)bigfastspace);
	vfree((void *)bighpispace);
	/*while (fast_stack) {
		onward = fast_stack->next;
		kfree_s((void *)fast_stack->data, sizeof(struct hit_packetinfo));
		kfree_s(fast_stack, sizeof(struct fast_timer_list));
		fast_stack = onward;
	}
	*/
}

void
re_fast_fill(void)
{
	if (!fast_stack)
		fast_fill(GFP_KERNEL);
}

struct fast_timer_list *
fast_alloc(int how)
{
	struct fast_timer_list *answer = fast_stack;

	if (answer) {
		fast_stack = answer->next;
		return answer;
	}
	fast_fill(how);
	answer = fast_stack;
	if (answer)
		fast_stack = answer->next;
	return answer;
}

void
fast_free(struct fast_timer_list *done)
{
	done->next = fast_stack;
	fast_stack = done;
}

/* Export interfaces */
static struct symbol_table nistnet_syms = {
#include <linux/symtab_begin.h>

	X(addmunge),
	X(rmmunge),

#include <linux/symtab_end.h>
};

void export_nistnet_symbols(void)
{
	register_symtab(&nistnet_syms);
}
